
package mapimporter;

/**
 *
 * @author David Fernandez
 */

import java.io.*;


public class CreateMap {

    static int MAX_PLAYERS = 64;
    char terrainMap[][];
    boolean riverMap[][];
    int startingPos[][];
    int gridx;
    int gridy;
    int topology;

    public CreateMap (String inCiv4, int riverOffset, boolean wrapEW, boolean wrapNS) throws Exception {
        String linea;
        String[] campo;
        int i = 0;

        int x = 0, y = 0;
        int team = 0;
        boolean isRiver = false;
        boolean isNofRiver = false;
        boolean isWofRiver = false;
        String featureType = "no";
        String terrainType = "no";
        String plotType = "no";

        char terrain;

        int fixedX = 0;
        int fixedY = 0;


        LineNumberReader inMap = null;        // Input Civ4 File
        try {
            inMap = new LineNumberReader(new FileReader(inCiv4));

            if (!wrapEW && !wrapNS) topology = 0;    //plane
            else if (wrapEW && !wrapNS) topology = 1;    //earth
            else if (!wrapEW && wrapNS) topology = 2;    //uranus
            else if (wrapEW && wrapNS) topology = 3;    //donut

            startingPos = new int[MAX_PLAYERS][2];
            for (i=0; i<MAX_PLAYERS; i++) {
                startingPos[i][0] = -1;
                startingPos[i][1] = -1;
            }
            // Read starting positions placed under "BeginPlayer"
            while ( ((linea = inMap.readLine()) != null ) && (linea.startsWith("BeginMap") == false) ) {
                linea=linea.trim();
                if (linea.startsWith("Team=")) {
                    team = Integer.valueOf(linea.substring(linea.indexOf("=")+1));
                }
                if (linea.startsWith("StartingX=")) {
                    campo = linea.split(",");
                    if (team < MAX_PLAYERS) {
                        campo[0] = campo[0].substring(linea.indexOf("=")+1);
                        startingPos[team][0] = Integer.valueOf(campo[0].trim());
                        campo[1] = campo[1].substring(linea.indexOf("=")+2);
                        startingPos[team][1] = Integer.valueOf(campo[1].trim());
                    }
                }
            }

            // Read map size placed under "BeginMap"
            while ( ((linea = inMap.readLine()) != null ) && (linea.startsWith("BeginPlot") == false) ) {
                linea=linea.trim();
                if (linea.startsWith("grid width=")) {
                    gridx = Integer.valueOf(linea.substring(linea.indexOf("=")+1));
                }
                if (linea.startsWith("grid height=")) {
                    gridy = Integer.valueOf(linea.substring(linea.indexOf("=")+1));
                }
            }
            if ( (gridx == 0) || (gridy == 0)) {
                throw new Exception ("ERROR: " + inCiv4 + " does not appear to be a valid Civ4 Map.");
            }

            terrainMap = new char[gridx][gridy];
            riverMap = new boolean[gridx][gridy];
            for (y=0; y<gridy; y++) {
                for (x=0; x<gridx; x++) {
                    terrainMap [x][y] = '+';
                    riverMap [x][y] = false;
                }
            }
            // Read terrain types and rivers placed under "BeginPlot"
            while ( ((linea = inMap.readLine()) != null ) ) {
                linea=linea.trim();
                if (linea.startsWith("x=")) {
                    campo = linea.split(",");
                    x = Integer.valueOf(campo[0].substring(linea.indexOf("=")+1));
                    y = Integer.valueOf(campo[1].substring(linea.indexOf("=")+1));//
                }
                if (linea.startsWith("isNOfRiver")) {
                    isRiver = true;
                    isNofRiver = true;
                }
                if (linea.startsWith("isWOfRiver")) {
                    isRiver = true;
                    isWofRiver = true;
                }
                if (linea.startsWith("FeatureType=")) {
                    campo = linea.split(",");
                    featureType = campo[0].substring(linea.indexOf("=")+1);
                }
                if (linea.startsWith("TerrainType=")) {
                    campo = linea.split(",");
                    terrainType = campo[0].substring(linea.indexOf("=")+1);
                }
                if (linea.startsWith("PlotType=")) {
                    campo = linea.split(",");
                    plotType = campo[0].substring(linea.indexOf("=")+1);
                }
                // Link each civ4 terrain type with freeciv char identifiers (rare lakes, and rare swamps)
                if (linea.startsWith("EndPlot")) {
                    if (plotType.equals("0")) terrain = 'm';
                    else if (plotType.equals("1")) terrain = 'h';
                    else if (plotType.equals("2") && terrainType.equals("TERRAIN_COAST")) terrain = '+';
                    else if (plotType.equals("3") && isRiver) terrain = 's';
                    else if (terrainType.equals("TERRAIN_SNOW")) terrain = 'a';
                    else if (terrainType.equals("TERRAIN_OCEAN")) terrain = ':';
                    else if (terrainType.equals("TERRAIN_COAST")) terrain = ' ';
                    else if (terrainType.equals("TERRAIN_DESERT")) terrain = 'd';
                    else if (featureType.equals("FEATURE_FOREST")) terrain = 'f';
                    else if (featureType.equals("FEATURE_JUNGLE")) terrain = 'j';
                    else if (terrainType.equals("TERRAIN_PLAINS")) terrain = 'p';
                    else if (terrainType.equals("TERRAIN_GRASS")) terrain = 'g';
                    else if (terrainType.equals("TERRAIN_TUNDRA")) terrain = 't';
                    else terrain = '+';

                    if ( (x < gridx) && (y >= 0) ) {
                        terrainMap[x][Y(y,gridy)] = terrain;

                        // Rivers placed in tile joints, adjusted to the center of the N-W tile by default
                        if (isRiver) {

                            switch (riverOffset) {
                                case 4: {   //S-E
                                    fixedX = x+1;
                                    fixedY = Y(y,gridy)+1;
                                    break;
                                }
                                case 3: {   //S-W
                                    fixedX = x;
                                    fixedY = Y(y,gridy)+1;
                                    break;
                                }
                                case 2: {   //N-E
                                    fixedX = x+1;
                                    fixedY = Y(y,gridy);
                                    break;
                                }
                                default: {   //N-W
                                    fixedX = x;
                                    fixedY = Y(y,gridy);
                                    break;
                                }
                            }
                            if (fixedY>=0 && fixedY<gridy && fixedX>=0 && fixedX<gridx) {
                                riverMap[fixedX][fixedY] = true;
                                if (isNofRiver && fixedX>=1) {
                                    // Needed to keep the same topology of the river while flowing through the tiles than it was in the joints
                                    riverMap[fixedX-1][fixedY] = true;
                                }
                            }
                        }
                    }

                    // Reset values for next loop, in case some entry is missing
                    x = 0; y = 0;
                    featureType = "no";
                    terrainType = "no";
                    plotType = "no";
                    isRiver = false;
                    isNofRiver = false;
                    isWofRiver = false;
                }
            }
        } catch (Exception e) {
            System.err.println(e.getLocalizedMessage());
        } finally {
            if (inMap != null) inMap.close();
        }
    }

    public void ExportMap (String inFreeciv, String outFreeciv, String[] scenarioNames) throws Exception {
        String linea;
        int x = 0, y = 0;
        int team = 0;

        LineNumberReader inScenario = null;   // Input Freeciv File
        PrintWriter outScenario = null;       // Output File

        try {
            if (inFreeciv.equals(outFreeciv)) {
                throw new Exception ("ERROR: Input and Output are the same file. Please, choose a different file to Save as");
            }
            inScenario = new LineNumberReader(new FileReader(inFreeciv));
            outScenario = new PrintWriter(new FileWriter(outFreeciv));

            // Read scenario options from Freeciv Scenario input
            while ( ((linea = inScenario.readLine()) != null ) && (linea.startsWith("[map]") == false) ) {

                if (linea.startsWith("name=_(\"")) {
                    outScenario.println("name=_(\"" + scenarioNames[0] + "\")");
                }
                else if (linea.startsWith("description=_(\"")) {
                    outScenario.println("description=_(\"" + scenarioNames[1] + "\")");
                }
                else if (linea.startsWith("metastring=\"")) {
                    outScenario.println("metastring=\"" + scenarioNames[2] + "\"");
                }
                else outScenario.println(linea);
            }
            if (linea == null) {
                throw new Exception ("ERROR: " + inFreeciv + " does not appear to be a valid Freeciv Scenario.");
            }

            outScenario.printf("[map]\n");
            outScenario.printf("topology_id=%d\n", topology);

            while ( ((linea = inScenario.readLine()) != null ) && (linea.startsWith("is_earth=") == false) ) {
                // Write the map sizes
                if (linea.startsWith("topology_id=")) {
                    // Topology does not appear in some scenarios
                }
                else if (linea.startsWith("width=")) {
                    outScenario.printf("width=%d\n", gridx);
                }
                else if (linea.startsWith("height=")) {
                    outScenario.printf("height=%d\n", gridy);
                }
                // Write the other map options
                else outScenario.println(linea);
            }
            if (linea == null) {
                throw new Exception ("ERROR: " + inFreeciv + " does not appear to be a valid Freeciv Scenario.");
            }

            outScenario.printf("is_earth=%d", topology);

            // Write starting positions
            for (team=0; (team<MAX_PLAYERS) && (startingPos[team][0]>=0); team++) {
                outScenario.printf( "\nr%dsx=%d", team, startingPos[team][0]);
                outScenario.printf( "\nr%dsy=%d", team, Y(startingPos[team][1],gridy));
            }

            // Write terrain types
            for (y=0; y<gridy; y++) {
                outScenario.printf( "\nt%03d=\"", y);
                for (x=0; x<gridx; x++) {
                    outScenario.print(terrainMap[x][y]);
                }
                outScenario.printf("\"");
            }

            // Write rivers as overlays
            for (y=0; y<gridy; y++) {
                outScenario.printf( "\nn%03d=\"", y);
                for (x=0; x<gridx; x++) {
                    if ((riverMap[x][y]) &&
                            (terrainMap[x][y] != ' ') &&    //no deep ocean
                            (terrainMap[x][y] != ':') &&    //no coast ocean
                            (terrainMap[x][y] != '+') )     //no lake
                    {
                        outScenario.print("2");
                    } else {
                        outScenario.print("0");
                    }
                }
                outScenario.printf("\"");
            }
            System.out.println("Successfully saved to: " + outFreeciv + ".");
        } catch (Exception e) {
            System.err.println(e.getLocalizedMessage());
        } finally {
            if (inScenario != null) inScenario.close();
            if (outScenario != null) outScenario.close();
        }
    }

    public int Y(int y, int gridy) {
        // Civ4 uses "y" coordinate from South to North, while in freeciv "Y" increases from North to South
        return (gridy-1-y);
    }

}